# -*- coding: utf-8 -*-
"""JsonSchema providers."""
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.interfaces import IASCII
from zope.schema.interfaces import IASCIILine
from zope.schema.interfaces import IBool
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IDecimal
from zope.schema.interfaces import IField
from zope.schema.interfaces import IFloat
from zope.schema.interfaces import IInt
from zope.schema.interfaces import IList
from zope.schema.interfaces import IObject
from zope.schema.interfaces import ISet
from zope.schema.interfaces import IText
from zope.schema.interfaces import ITextLine
from zope.schema.interfaces import ITuple
from zope.schema.interfaces import IVocabularyFactory

from plone.restapi.types.interfaces import IJsonSchemaProvider
from plone.restapi.types.utils import get_fields_from_schema


@adapter(IField, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DefaultJsonSchemaProvider(object):

    def __init__(self, field, context, request):
        self.field = field
        self.context = context
        self.request = request

    def additional(self):
        """Additional info for field."""
        return {}

    def get_title(self):
        return translate(self.field.title, context=self.request)

    def get_description(self):
        if self.field.description is None:
            return u''

        return translate(self.field.description, context=self.request)

    def get_schema(self):
        """Get jsonschema for field.

        You should override `additional` method to provide more properties
        about the field."""
        schema = {
            'type': self.get_type(),
            'title': self.get_title(),
            'description': self.get_description(),
        }
        if self.field.default is not None:
            schema['default'] = self.field.default

        schema.update(self.additional())
        return schema

    def get_type(self):
        raise NotImplementedError


@adapter(ITextLine, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TextLineJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'string'


@adapter(IText, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TextJsonSchemaProvider(TextLineJsonSchemaProvider):

    def additional(self):
        info = {}
        if self.field.min_length is not None:
            info['minLength'] = self.field.min_length

        if self.field.max_length is not None:
            info['maxLength'] = self.field.max_length

        return info


@adapter(IASCII, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ASCIIJsonSchemaProvider(TextLineJsonSchemaProvider):

    pass


@adapter(IASCIILine, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ASCIILineJsonSchemaProvider(TextLineJsonSchemaProvider):

    pass


@adapter(IFloat, Interface, Interface)
@implementer(IJsonSchemaProvider)
class FloatJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'number'

    def additional(self):
        info = {}
        if self.field.min is not None:
            info['minimum'] = self.field.min

        if self.field.max is not None:
            info['maximum'] = self.field.max

        return info


@adapter(IDecimal, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DecimalJsonSchemaProvider(FloatJsonSchemaProvider):

    pass


@adapter(IInt, Interface, Interface)
@implementer(IJsonSchemaProvider)
class IntegerJsonSchemaProvider(FloatJsonSchemaProvider):

    def get_type(self):
        return 'integer'


@adapter(IBool, Interface, Interface)
@implementer(IJsonSchemaProvider)
class BoolJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'boolean'


@adapter(ICollection, Interface, Interface)
@implementer(IJsonSchemaProvider)
class CollectionJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'array'

    def get_items(self):
        """Get items properties."""
        value_type_adapter = getMultiAdapter(
            (self.field.value_type, self.context, self.request),
            IJsonSchemaProvider)

        return value_type_adapter.get_schema()

    def additional(self):
        info = {}
        info['additionalItems'] = True
        if self.field.min_length:
            info['minItems'] = self.field.min_length

        if self.field.max_length:
            info['maxItems'] = self.field.max_length

        info['items'] = self.get_items()

        return info


@adapter(IList, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ListJsonSchemaProvider(CollectionJsonSchemaProvider):

    def additional(self):
        info = super(ListJsonSchemaProvider, self).additional()
        if IChoice.providedBy(self.field.value_type):
            info['uniqueItems'] = True
        else:
            info['uniqueItems'] = False

        return info


@adapter(ISet, Interface, Interface)
@implementer(IJsonSchemaProvider)
class SetJsonSchemaProvider(CollectionJsonSchemaProvider):

    def additional(self):
        info = super(SetJsonSchemaProvider, self).additional()
        info['uniqueItems'] = True
        return info


@adapter(ITuple, Interface, Interface)
@implementer(IJsonSchemaProvider)
class TupleJsonSchemaProvider(SetJsonSchemaProvider):

    pass


@adapter(IChoice, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ChoiceJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'string'

    def additional(self):
        # choices and enumNames are v5 proposals, for now we implement both
        choices = []
        enum = []
        enum_names = []
        if self.field.vocabularyName:
            vocabulary = getUtility(
                IVocabularyFactory,
                name=self.field.vocabularyName)(self.context)
        else:
            vocabulary = self.field.vocabulary

        if hasattr(vocabulary, '__iter__'):
            for term in vocabulary:
                title = translate(term.title, context=self.request)
                choices.append((term.token, title))
                enum.append(term.token)
                enum_names.append(title)

            return {
                'enum': enum,
                'enumNames': enum_names,
                'choices': choices,
            }
        else:
            return {}


@adapter(IObject, Interface, Interface)
@implementer(IJsonSchemaProvider)
class ObjectJsonSchemaProvider(DefaultJsonSchemaProvider):

    prefix = ''

    def get_type(self):
        return 'object'

    def get_properties(self):
        if self.prefix:
            prefix = '.'.join([self.prefix, self.field.__name__])
        else:
            prefix = self.field.__name__

        return get_fields_from_schema(
            self.field.schema, self.context, self.request, prefix)

    def additional(self):
        info = super(ObjectJsonSchemaProvider, self).additional()
        info['properties'] = self.get_properties()
        return info


@adapter(IDate, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DateJsonSchemaProvider(DefaultJsonSchemaProvider):

    def get_type(self):
        return 'string'

    def additional(self):
        info = {}
        if self.field.min is not None:
            info['minimum'] = self.field.min

        if self.field.max is not None:
            info['maximum'] = self.field.max

        return info


@adapter(IDatetime, Interface, Interface)
@implementer(IJsonSchemaProvider)
class DatetimeJsonSchemaProvider(DateJsonSchemaProvider):

    pass



# -*- coding: utf-8 -*-
"""Utils for jsonschema."""
from collections import OrderedDict

from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.schema import getFieldsInOrder

from plone.autoform.interfaces import IFormFieldProvider
from plone.behavior.interfaces import IBehavior
from plone.supermodel.interfaces import FIELDSETS_KEY
from Products.CMFCore.utils import getToolByName

from plone.restapi.types.interfaces import IJsonSchemaProvider


def non_fieldset_fields(schema):
    fieldset_fields = []
    fieldsets = schema.queryTaggedValue(FIELDSETS_KEY, [])

    for fieldset in fieldsets:
        fieldset_fields.extend(fieldset.fields)

    fields = [info[0] for info in getFieldsInOrder(schema)]
    return [f for f in fields if f not in fieldset_fields]


def get_ordered_fields(fti):
    # this code is much complicated because we have to get sure
    # we get the fields in the order of the fieldsets
    # the order of the fields in the fieldsets can differ
    # of the getFieldsInOrder(schema) order...
    # that's because fields from different schemas
    # can take place in the same fieldset
    schema = fti.lookupSchema()
    fieldset_fields = {}
    ordered_fieldsets = ['default']
    for fieldset in schema.queryTaggedValue(FIELDSETS_KEY, []):
        ordered_fieldsets.append(fieldset.__name__)
        fieldset_fields[fieldset.__name__] = fieldset.fields

    fieldset_fields['default'] = non_fieldset_fields(schema)

    # Get the behavior fields
    fields = getFieldsInOrder(schema)
    for behavior_id in fti.behaviors:
        schema = getUtility(IBehavior, behavior_id).interface
        if not IFormFieldProvider.providedBy(schema):
            continue

        fields.extend(getFieldsInOrder(schema))
        for fieldset in schema.queryTaggedValue(FIELDSETS_KEY, []):
            fieldset_fields.setdefault(fieldset.__name__, []).extend(
                fieldset.fields)
            ordered_fieldsets.append(fieldset.__name__)

        fieldset_fields['default'].extend(non_fieldset_fields(schema))

    ordered_fields = []
    for fieldset in ordered_fieldsets:
        ordered_fields.extend(fieldset_fields[fieldset])

    fields.sort(key=lambda field: ordered_fields.index(field[0]))
    return fields


def get_fields_from_schema(schema, context, request, prefix='',
                           excluded_fields=None):
    """Get jsonschema from zope schema."""
    fields_info = OrderedDict()
    if excluded_fields is None:
        excluded_fields = []

    for fieldname, field in getFieldsInOrder(schema):
        if fieldname not in excluded_fields:
            adapter = getMultiAdapter(
                (field, context, request),
                interface=IJsonSchemaProvider)

            adapter.prefix = prefix
            if prefix:
                fieldname = '.'.join([prefix, fieldname])

            fields_info[fieldname] = adapter.get_schema()

    return fields_info


def get_jsonschema_for_fti(fti, context, request, excluded_fields=None):
    """Get jsonschema for given fti."""
    fields_info = OrderedDict()
    if excluded_fields is None:
        excluded_fields = []

    required = []
    for fieldname, field in get_ordered_fields(fti):
        if fieldname not in excluded_fields:
            adapter = getMultiAdapter(
                (field, context, request),
                interface=IJsonSchemaProvider)
            # get name from z3c.form field to have full name (behavior)
            fields_info[fieldname] = adapter.get_schema()
            if field.required:
                required.append(fieldname)

    return {
        'type': 'object',
        'title': fti.Title(),
        'properties': fields_info,
        'required': required,
    }


def get_jsonschema_for_portal_type(portal_type, context, request,
                                   excluded_fields=None):
    """Get jsonschema for given portal type name."""
    ttool = getToolByName(context, 'portal_types')
    fti = ttool[portal_type]
    return get_jsonschema_for_fti(
        fti, context, request, excluded_fields=excluded_fields)



# -*- coding: utf-8 -*-
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone.jsonserializer.interfaces import IJsonCompatible
from zope.component.hooks import getSite
from zope.component import adapter
from zope.i18nmessageid.message import Message
from zope.interface import implementer
from zope.interface import Interface

try:
    from plone.app.textfield import IRichTextValue
    HAS_RICH_TEXT_VALUE = True
except ImportError:
    HAS_RICH_TEXT_VALUE = False

try:
    import Missing
    HAS_ZOPE_MISSING = True
except ImportError:
    HAS_ZOPE_MISSING = False

try:
    unicode
except NameError:
    unicode = str

try:
    long
except NameError:
    long = int

try:
    from Products.CMFPlone.utils import safe_unicode
except ImportError:
    safe_unicode = str

try:
    from Products.CMFPlone.utils import getSiteEncoding
except ImportError:
    HAS_SITE = False
else:
    HAS_SITE = True

try:
    from DateTime import DateTime
except ImportError:
    HAS_ZOPE_DATETIME = False
else:
    HAS_ZOPE_DATETIME = True


def json_compatible(value):
    """The json_compatible function converts any value to JSON compatible
    data when possible, raising a TypeError for unsupported values.
    This is done by using the IJsonCompatible converters.

    Be aware that adapting the value `None` will result in a component
    lookup error unless `None` is passed in as default value.
    Because of that the `json_compatible` helper method should always be
    used for converting values that may be None.
    """
    return IJsonCompatible(value, None)


def encoding():
    if HAS_SITE:
        return getSiteEncoding(getSite())
    else:
        return 'utf-8'


@adapter(Interface)
@implementer(IJsonCompatible)
def default_converter(value):
    if value is None:
        return value

    if type(value) in (unicode, bool, int, float, long):
        return value

    raise TypeError(
        'No converter for making'
        ' {0!r} ({1}) JSON compatible.'.format(value, type(value)))


@adapter(str)
@implementer(IJsonCompatible)
def string_converter(value):
    return safe_unicode(value, )


@adapter(list)
@implementer(IJsonCompatible)
def list_converter(value):
    return list(map(json_compatible, value))


@adapter(PersistentList)
@implementer(IJsonCompatible)
def persistent_list_converter(value):
    return list_converter(value)


@adapter(tuple)
@implementer(IJsonCompatible)
def tuple_converter(value):
    return list(map(json_compatible, value))


@adapter(frozenset)
@implementer(IJsonCompatible)
def frozenset_converter(value):
    return list(map(json_compatible, value))


@adapter(set)
@implementer(IJsonCompatible)
def set_converter(value):
    return list(map(json_compatible, value))


@adapter(dict)
@implementer(IJsonCompatible)
def dict_converter(value):
    if value == {}:
        return {}

    keys, values = zip(*value.items())
    keys = map(json_compatible, keys)
    values = map(json_compatible, values)
    return dict(zip(keys, values))


@adapter(PersistentMapping)
@implementer(IJsonCompatible)
def persistent_mapping_converter(value):
    return dict_converter(value)


@adapter(datetime)
@implementer(IJsonCompatible)
def python_datetime_converter(value):
    if HAS_ZOPE_DATETIME:
        return json_compatible(DateTime(value))
    else:
        return json_compatible(value.isoformat())


if HAS_ZOPE_DATETIME:
    @adapter(DateTime)
    @implementer(IJsonCompatible)
    def zope_DateTime_converter(value):
        return json_compatible(value.ISO8601())


@adapter(date)
@implementer(IJsonCompatible)
def date_converter(value):
    return json_compatible(value.isoformat())


@adapter(time)
@implementer(IJsonCompatible)
def time_converter(value):
    return json_compatible(value.isoformat())


@adapter(timedelta)
@implementer(IJsonCompatible)
def timedelta_converter(value):
    return json_compatible(value.total_seconds())


if HAS_RICH_TEXT_VALUE:
    @adapter(IRichTextValue)
    @implementer(IJsonCompatible)
    def richtext_converter(value):
        return {
            u'data': json_compatible(value.raw),
            u'content-type': json_compatible(value.mimeType),
            u'output-content-type': json_compatible(value.outputMimeType),
            u'encoding': json_compatible(value.encoding),
        }


@adapter(Message)
@implementer(IJsonCompatible)
def i18n_message_converter(value):
    # TODO:
    # value = translate(value, context=getRequest())
    return value


if HAS_ZOPE_MISSING:
    @adapter(Missing.Value.__class__)
    @implementer(IJsonCompatible)
    def missing_value_converter(value):
        return None



# -*- coding: utf-8 -*-
from plone.jsonserializer.interfaces import ISchemaCompatible
from zope.component import adapter
from zope.component import ComponentLookupError
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface
from zope.schema._bootstrapinterfaces import IFromUnicode
from zope.schema.interfaces import IDict
from zope.schema.interfaces import IBool
from zope.schema.interfaces import IField
from zope.schema.interfaces import IFrozenSet
from zope.schema.interfaces import IList
from zope.schema.interfaces import ISet
from zope.schema.interfaces import ITuple
import logging

try:
    from plone.app.textfield import IRichText
    from plone.app.textfield import RichTextValue
    HAS_RICH_TEXT_VALUE = True
except ImportError:
    HAS_RICH_TEXT_VALUE = False

logger = logging.getLogger('plone.jsonserializer')


def schema_compatible(value, schema_or_field):
    """The schema_compatible function converts any value to zope.schema
    compatible data when possible, raising a TypeError for unsupported values.
    This is done by using the ISchemaCompatible converters.
    """
    if value is None:
        return value

    try:
        return getMultiAdapter((value, schema_or_field), ISchemaCompatible)
    except ComponentLookupError:
        logger.error((u'Deserializer not found for field type '
                      u'"{0:s}" with value "{1:s}" and it was '
                      u'deserialized to None.').format(
            schema_or_field, value))
        return None


@adapter(dict, Interface)
@implementer(ISchemaCompatible)
def schema_dict_converter(value, schema):
    if value == {}:
        return {}

    items = [(k, v) for k, v in value.items() if k in schema]
    keys, values = zip(*items)
    keys = map(str, keys)
    values = [schema_compatible(values[idx], schema[keys[idx]])
              for idx in range(len(keys))]
    return dict(zip(keys, values))


@adapter(Interface, IField)
@implementer(ISchemaCompatible)
def default_converter(value, field):
    return value


@adapter(Interface, IBool)
@implementer(ISchemaCompatible)
def bool_converter(value, field):
    return bool(value)


@adapter(Interface, IFromUnicode)
@implementer(ISchemaCompatible)
def from_unicode_converter(value, field):
    try:
        return field.fromUnicode(value)
    except UnicodeEncodeError:
        return value.encode('utf-8', 'ignore')


@adapter(list, IList)
@implementer(ISchemaCompatible)
def list_converter(value, field):
    return [schema_compatible(item, field.value_type)
            for item in value]


@adapter(list, ITuple)
@implementer(ISchemaCompatible)
def tuple_converter(value, field):
    return tuple(list_converter(value, field))


@adapter(list, ISet)
@implementer(ISchemaCompatible)
def set_converter(value, field):
    return set(list_converter(value, field))


@adapter(list, IFrozenSet)
@implementer(ISchemaCompatible)
def frozenset_converter(value, field):
    return frozenset(list_converter(value, field))


@adapter(dict, IDict)
@implementer(ISchemaCompatible)
def dict_converter(value, field):
    if value == {}:
        return {}


    keys, values = zip(*value.items())
    keys = [schema_compatible(keys[idx], field.key_type)
            for idx in range(len(keys))]
    values = [schema_compatible(values[idx], field.value_type)
              for idx in range(len(values))]
    value = dict(zip(keys, values))

    return value


if HAS_RICH_TEXT_VALUE:
    @adapter(dict, IRichText)
    @implementer(ISchemaCompatible)
    def richtext_converter(value, schema):
        encoding = value.get('encoding', u'utf-8')\
                        .encode('utf-8', 'ignore')
        raw = value.get('data', '').encode(encoding)
        mimeType = value.get('content-type', u'text/html')\
                        .encode('utf-8', 'ignore')
        outputMimeType = value.get('output-content-type', u'text/x-safe-html')\
                              .encode('utf-8', 'ignore')
        return RichTextValue(
            raw=raw,
            mimeType=mimeType,
            outputMimeType=outputMimeType,
            encoding=encoding
        )



# -*- coding: utf-8 -*-
from plone.jsonserializer.exceptions import DeserializationError

import json


def json_body(request):
    try:
        data = json.loads(request.get('BODY', '{}'))
    except ValueError:
        raise DeserializationError('No JSON object could be decoded')
    if not isinstance(data, dict):
        raise DeserializationError('Malformed body')
    return data



# -*- coding: utf-8 -*-


class DeserializationError(Exception):
    """An error happened during deserialization of content.
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class QueryParsingError(Exception):
    """An error happened while parsing a search query.
    """



from setuptools import setup, find_packages

version = '0.9.4'

long_description = (
    open('README.md').read() + '\n' +
    open('CHANGES.md').read() + '\n'
)

setup(name='plone.jsonserializer',
      version=version,
      description="JSON serialization/deserialization adapters for Plone.",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Environment :: Web Environment",
          "Framework :: Plone",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      keywords='',
      author='Plone Foundation',
      author_email='foundation@plone.org',
      url='https://github.com/pyrenees/plone.jsonserializer/',
      license='gpl',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['plone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
      ],
      extras_require={
          'test': [
              'plone.app.testing >= 4.2.2',
          ]},
      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



# -*- coding: utf-8 -*-

# pylint: disable=E0211, W0221
# E0211: Method has no argument
# W0221: Arguments number differs from overridden '__call__' method

from zope.interface import Interface


class ISerializeToJson(Interface):
    """Adapter to serialize a Dexterity object into a JSON object.
    """


class ISerializeToJsonSummary(Interface):
    """Adapter to serialize an object into a JSON compatible summary that
    contains only the most basic information.
    """


class IJsonCompatible(Interface):
    """Convert a value to a JSON compatible data structure.
    """


class ISchemaCompatible(Interface):
    """Convert a value to a zope schema ompatible data structure.
    """


class ISchemaSerializer(Interface):
    """ The fieldset serializer multi adapter serializes the schema
    into JSON compatible python data.
    """


class IFieldsetSerializer(Interface):
    """The fieldset serializer multi adapter serializes the fieldset and its fields
    into JSON compatible python data.
    """

    def __init__(fieldset, context, request):
        """Adapts fieldset, context and request.
        """

    def __call__():
        """Returns JSON compatible python data.
        """


class IFieldSerializer(Interface):
    """The field serializer multi adapter serializes the field value into
    JSON compatible python data.
    """

    def __init__(field, context, request):
        """Adapts field, context and request.
        """

    def __call__():
        """Returns JSON compatible python data.
        """


class IDeserializeFromJson(Interface):
    """An adapter to deserialize a JSON object into an object in Plone."""


class IFieldDeserializer(Interface):
    """An adapter to deserialize a JSON value into a field value.
    """

    def __init__(field, context, request):
        """Adapts a field, it's context and the request.
        """

    def __call__(value):
        """Convert the provided JSON value to a field value.
        """
